;  Emulation for 65816 Microprocessor
;  (C) 1997 Robert William Grubbs, All Rights Reserved
;  Written for Turbo Assembler 1.0
;  Latest revision 9/21/97

   .MODEL compact        ;using simplified segments for now
   .STACK 0400h
   .DATA
   .386
   NOJUMPS

;Emulated Registers
eDB db 0       ;Data Bank register
eXL db 0       ;Index Register (X) low
eXH db 0       ;Index Register (X) high
eYL db 0       ;Index Register (Y) low
eYH db 0       ;Index Register (Y) high
eSL db 0       ;Stack Pointer (S) low
eSH db 0       ;Stack Pointer (S) high
eAL db 0       ;Accumulator (A) low
eAH db 0       ;Accumulator (A) high
ePB db 0       ;Program Bank (PB)
ePCL db 0      ;Program Counter (PC) low
ePCH db 0      ;Program Counter (PC) high
eDL db 0       ;Direct Register (D) low
eDH db 0       ;Direct Register (D) high
               ;Processor Status Register (P)
ePn db 0       ; negative flag
ePv db 0       ; overflow flag
ePm db 0       ; accumulator size select flag
ePx db 0       ; index size select flag
ePd db 0       ; decimal flag
ePi db 0       ; interrupt flag
ePz db 0       ; zero flag
ePc db 0       ; carry flag
ePE db 0       ;Processor Status Register (Extended)
eSTAT db 0     ;Emulation Status (for WAI, STP instructions)
eREND db 0     ;Render Trigger (set by $4210)      [Obsoleted]
eSS1 db 0      ;enable sound skip method #1 (force zero flag high two cycles)
eJDebug dw 0   ;Break address
eJDebugB db 0  ;Break Bank
eJDebugOn db 0 ;enable jump debug trace over
eIRQFlag db 0  ;was IRQ already pressed?
eINSTPC dd 15000  ;instructions to execute between NMI calls
eNMICount dd 0 ;running count of NMI frames

;Internal Emulation Data
eEX dd 0       ;Total Ops executed (for NMI trigger)

   .CODE
EmuStart:                ;the equivilent of main()
  NOJUMPS                ;disable auto-jumpsizing (since it never works right)
  mov ax,@DATA
  mov ds,ax              ;get the data segment

  call SNESRomInit       ;initialize the memory, and load DEBUG.SMC
  cmp CS:[mERROR],0      ;check for allocation error
  jne EmuEnd             ;on error, end the emulation

  IFDEF VIDEOON
    mov ax,0013h         ;
    int 10h              ;Enter graphics mode
    call TweakVideo      ;Straigt from NESA (converted for TASM)
  ENDIF

  call ResetCPU          ;"hardware" reset of the e65816

  EmuLoop:               ;main emulation loop (for now)
  cmp [eJDebugOn],0           ;Debugging crap  (breakpoints)
  je NoJDebug
  mov ax,WORD PTR [eJDebug]
  cmp WORD PTR [ePCL],ax
  jne NoJDebug
  mov al,[eJDebugB]
  cmp [ePB],al
  jne NoJDebug
  mov WORD PTR CS:[dWC],0
  NoJDebug:
  mov eax,[eINSTPC]
  cmp [eEX],eax          ; generate NMI every 15,000 instructions
  jne SHORT SkipNMISet
  IFDEF VIDEOON
    call vModeX          ; If compiled in Video mode, render the screen
  ENDIF
  mov [eEX],0
  test [r4200],80h       ; Test for NMI Enable
  jz SkipNMISet
  inc [eNMICount]        ; Count NMIs (for debugging)
  call eNMI              ; call NMI
  SkipNMISet:
  inc DWORD PTR [eEX]    ;count instructions
  cmp [eSS1],0           ;Test for Sound Skip mode
  je NoSS1
  dec [eSS1]             ;decrement SS cycle count
  mov [ePz],2            ;Force Z flag high
  NoSS1:

  mov al,BYTE PTR CS:[r2100]   ;Create some convienient breakpoints
  cmp al,1
  jne DebugPointA              ;Trap end of "Fade out" in MARIO.SMC
  nop
  DebugPointA:
  test [ePd],8h
  jz DebugPointB               ;Trap Decimal flag being set
  nop
  DebugPointB:
  cmp [ePx],0
  je NoKillXRegs     ;Tossed this in to see if it would fix the big bug
  mov [eXH],0
  mov [eYH],0
  NoKillXRegs:



  call ExecCPU           ;execute a e65816 instruction
  IFDEF VIDEOON
    mov ah,02                  ;Test for keypresses in Video mode
    int 16h
    test al,1h                 ;Shift keys activate wait for keypress mode
    jnz NoResetIRQFlag
    mov [eIRQFlag],0           ;On RIGHT shift, force IRQ
    NoResetIRQFlag:
    and al,0fh
    test al,1h
    jz NoIRQ
    cmp [eIRQFlag],0
    jne SHORT NoIRQ
    mov [eIRQFlag],1
    call eIRQ
    xor al,al
    jmp SHORT RendTestS
    NoIRQ:
    or al,al
    jz RendTestS
    mov ah,0
    int 16h
    cmp al,'q'                  ;Quit
    jne RendTestSA
    mov [eSTAT],1
    RendTestSA:
    cmp al,'1'
    jne RendTestSB
    and [r212C],11111110b       ;Disable BG1
    RendTestSB:
    cmp al,'2'
    jne RendTestSC
    and [r212C],11111101b       ;Disable BG2
    RendTestSC:
    cmp al,'3'
    jne RendTestSD
    and [r212C],11111011b       ;Disable BG3
    RendTestSD:
    cmp al,'4'
    jne RendTestSE
    and [r212C],11110111b       ;Disable BG4
    RendTestSE:
    cmp al,'n'
    jne RendTestS
    call eNMI                   ;Force NMI
    RendTestS:
  ENDIF
  cmp [eSTAT],0          ;check emulation status
  jz EmuLoop             ;Loop back to next emulation loop
  IFDEF VIDEOON
    mov [r212C],1h
    call vModeX          ;Force render before quitting, and pause
    mov ah,00
    int 16h
    mov [r212C],2h
    call vModeX          ;Force render before quitting, and pause
    mov ah,00
    int 16h
    mov [r212C],4h
    call vModeX          ;Force render before quitting, and pause
    mov ah,00
    int 16h
    mov [r212C],8h
    call vModeX          ;Force render before quitting, and pause
    mov ah,00
    int 16h
  ENDIF
EmuEnd:
  call SNESRomKill       ;deallocate RAM

  IFDEF VIDEOON
    mov ax,0003h
    int 10h              ;Switch back to text mode
  ENDIF
  mov ah,04ch
  int 21h                 ;return to DOS


MemRead EQU xMemRead      ;redefine MemRead to xMemRead
MemWrite EQU xMemWrite    ;redefine MemWrite to xMemWrite

PushAL MACRO              ;eStack push macro
  LOCAL @@PushALS
  xor ah,ah
  mov bx,WORD PTR [eSL]
  call MemWrite
  dec WORD PTR [eSL]
  cmp [ePE],0
  je @@PushALS
  mov [eSH],1
  @@PushALS:
ENDM
PopAL MACRO               ;eStack pop macro (no overflow checking)
  mov ah,0
  mov bx,WORD PTR [eSL]
  call MemRead
  dec WORD PTR [eSL]
ENDM
Test16Bit MACRO           ;Obsoleted macro, but in case I still use it...
  LOCAL @@Test16BitS
  xor dh,dh
  cmp [ePE],1
  je SHORT @@Test16BitS
  inc bx
  call MemRead
  mov dh,al
  @@Test16BitS:
ENDM
Immediate PROC            ;Return immediate (8/16 bit) value in DX,carry
  xor dh,dh                 ;init high byte of output
  mov ah,[ePB]              ;get the program bank
  mov bx,WORD PTR [ePCL]    ;get the program counter
  call MemRead              ;get low byte of immediate value
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov dl,al                 ;save low byte into output (DX)
  test [ePm],20h             ;is 16bit mode?
  jnz SHORT @@ImmediateS          ;if not, we're done
  inc bx                    ;increment internal memory pointer
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  call MemRead              ;get high byte of immediate value
  mov dh,al                 ;store high byte into output (DX)
  stc                       ;mark 16bit mode
  @@ImmediateS:             ;
  ret                       ;
ENDP
eIMM MACRO
  call Immediate
ENDM
PullStack PROC            ;eStack pull  (why do I have 2?  <sigh>)
  mov bx,WORD PTR [eSL]     ;get stack pointer
  inc bx                    ;increment it
  cmp [ePE],0               ;test for emulation mode
  je PullStackS             ;
  mov bh,1                  ;in emulation mode, fix overrun problem
  PullStackS:               ;
  mov WORD PTR [eSL],bx     ;store updated stack pointer
  xor ah,ah                 ;clear ah
  call MemRead              ;get data off of stack
  ret                       ;
ENDP
ePOP MACRO
  call PullStack
ENDM
Immediate8 PROC            ;Return immediate (8-bit only) value in DL
  mov ah,[ePB]              ;get the program bank
  mov bx,WORD PTR [ePCL]    ;get the program counter
  call MemRead              ;get low byte of immediate value
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov dl,al                 ;save low byte into output (DX)
  ret                       ;
ENDP
eIMM8 MACRO
  call Immediate8
ENDM
ImmediateX PROC            ;Return immediate (8/16 bit) value in DX,carry
  xor dh,dh                 ;init high byte of output
  mov ah,[ePB]              ;get the program bank
  mov bx,WORD PTR [ePCL]    ;get the program counter
  call MemRead              ;get low byte of immediate value
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov dl,al                 ;save low byte into output (DX)
  test [ePx],10h             ;is 16bit mode?
  jnz SHORT @@ImmediateXS          ;if not, we're done
  inc bx                    ;increment internal memory pointer
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  call MemRead              ;get high byte of immediate value
  mov dh,al                 ;store high byte into output (DX)
  stc                       ;mark 16bit mode
  @@ImmediateXS:             ;
  ret                       ;
ENDP
eIMMX MACRO
  call ImmediateX
ENDM
Accumulator PROC          ;Return Accumulator (8/16 bit) value in DX,carry
  xor dh,dh                 ;clear upper 8 bits of output
  mov dl,[eAL]              ;get lower 8 bits into output (DL)
  test [ePm],20h            ;test for 16-bit mode, clear carry
  jnz SHORT @@AccumulatorS  ;
  mov dh,[eAH]              ;get upper 8 bits into output (DH)
  stc                       ;mark 16-bit value
  @@AccumulatorS:           ;
  ret                       ;
ENDP
eACC MACRO
  call Accumulator
ENDM
Relative PROC
  mov ah,[ePB]              ;get the program bank
  mov bx,WORD PTR [ePCL]    ;get the program counter
  call MemRead              ;get low byte of immediate value
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  cbw                       ;convert signed byte into signed word
  mov dx,ax                 ;save low byte into output (DX)
  ret
ENDP
eREL MACRO
  call Relative
ENDM
Absolute PROC             ;Set AH:BX to 24-bit effective address, Absolute mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;read AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal memory pointer
  call MemRead              ;read AAH
  mov dh,al                 ;preserve AAH
  inc WORD PTR [ePCL]       ;
  inc WORD PTR [ePCL]       ;increment program counter +2
  inc ch                    ;
  inc ch                    ;increment cycle count +2
  mov ah,[eDB]              ;read data bank
  mov bx,dx                 ;set output (BX) to low 16-bits of address
  ret                       ;
ENDP
eABS MACRO
  call Absolute
ENDM
AbsoluteLong PROC         ;Set AH:BX to 24-bit effective address, Absolute Long mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR[ePCL]     ;get program counter
  call MemRead              ;read AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal memory pointer
  call MemRead              ;read AAH
  mov dh,al                 ;preserve AAH
  add WORD PTR [ePCL],3     ;increment Program Counter +3
  add ch,3                  ;increment cycle count +3
  inc bx                    ;increment internal memory pointer
  call MemRead              ;read AAB
  mov ah,al                 ;set output (AH) to high 8-bits of address
  mov bx,dx                 ;set output (BX) to low 16-bits of address
  ret                       ;
ENDP
eABL MACRO
  call AbsoluteLong
ENDM
Direct PROC               ;Set AH:BX to 24-bit effective address, Direct mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get progam counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset into output (BX)
  cmp [eDL],0               ;is DL page aligned?
  je SHORT @@SkipDirect           ;yes, skip next instruction
  inc ch                    ;increment cycle count
  @@SkipDirect:             ;
  ret                       ;
ENDP
eDIR MACRO
  call Direct
ENDM
DirectIndirectIndexed PROC ;Set AH:BX to 24-bit effective address, Direct-Indirect-Indexed mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DIISkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DIISkip3:               ;
  call MemRead              ;get AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal memory pointer
  add cl,3                  ;increment cycle count
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  mov bx,WORD PTR [eYL]     ;get Y-index
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@DIISkip              ;is 16bit, skip next instruction
  dec ch                    ;decrement cycle count
  xor bh,bh                 ;clear YH
  @@DIISkip:                ;
  add bx,dx                 ;add memory pointer into output (BX)
  mov ah,[eDB]              ;get data bank into output (AH)
  adc ah,0                  ;increment data bank if necessary
  ret                       ;
ENDP
eDII MACRO
  call DirectIndirectIndexed
ENDM
DirectIndirectIndexedLong PROC    ;
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DXLSkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DXLSkip3:               ;
  call MemRead              ;get AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal memory pointer
  add cl,3                  ;increment cycle count +3
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  inc bx                    ;increment internal memory pointer
  call MemRead              ;get AAB
  mov ah,al                 ;set output (AH) to AAB
  mov bx,WORD PTR [eYL]     ;get Y-index
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@DXLSkip              ;is 16bit, skip next instruction
  xor bh,bh                 ;clear YH
  @@DXLSkip:                ;
  add bx,dx                 ;add memory pointer into output (BX)
  adc ah,0                  ;increment data bank if necessary
  ret                       ;
ENDP
eDXL MACRO
  call DirectIndirectIndexedLong
ENDM
DirectIndexedIndirect PROC    ;
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR[ePCL]     ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  add ch,4                  ;increment cycle count +4
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DXISkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DXISkip3:               ;
  mov ax,WORD PTR [eXL]     ;get X index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@DXISkip              ;is 16bit, skip next instruction
  xor ah,ah                 ;clear XH
  @@DXISkip:                ;
  add bx,ax                 ;add X index register
  xor ah,ah                 ;clear ah
  call MemRead              ;get AAL
  inc bx                    ;increment internal pointer
  mov dl,al                 ;preserve AAL
  call MemRead              ;get AAH
  mov bh,al                 ;set output (BH) to middle 8-bits of address
  mov bl,dl                 ;set output (BL) to low 8-bits of address
  mov ah,[eDB]              ;set output (AH) to data bank
  ret                       ;
ENDP
eDXI MACRO
  call DirectIndexedIndirect
ENDM
DirectIndexedX PROC       ;Set AH:BX to 24-bit effective address, Direct-Indexed-X mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DXXSkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DXXSkip3:               ;
  mov ax,WORD PTR [eXL]     ;get X-index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@DXXSkip              ;is 16bit, skip next instruction
  xor ah,ah                 ;clear XH
  @@DXXSkip:                ;
  add bx,ax                 ;add X-index
  inc ch                    ;increment cycle count
  xor ah,ah                 ;clear ah
  ret                       ;note bank wraps, unlike absolute addressing
ENDP
eDXX MACRO
  call DirectIndexedX
ENDM
DirectIndexedY PROC       ;Set AH:BX to 24-bit effective address, Direct-Indexed-Y mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR[ePCL]     ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DXYSkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DXYSkip3:               ;
  mov ax,WORD PTR [eYL]     ;get Y-index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@DXYSkip              ;is 16bit, skip next instruction
  xor ah,ah                 ;clear XH
  @@DXYSkip:                ;
  add bx,ax                 ;add Y-index
  inc ch                    ;increment cycle count
  xor ah,ah                 ;clear ah
  ret                       ;note bank wraps, unlike absolute addressing
ENDP
eDXY MACRO
  call DirectIndexedY
ENDM
AbsoluteIndexX PROC       ;Set AH:BX to 24-bit effective address, Absolute-Indexed-X mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR[ePCL]     ;get program counter
  call MemRead              ;get AAL
  add ch,3                  ;increment cycle count +3
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  inc WORD PTR [ePCL]       ;
  inc WORD PTR [ePCL]       ;increment program counter +2
  mov ah,[eDB]              ;get data bank register
  mov bx,WORD PTR [eXL]     ;get X index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@AXXSkip              ;is 16bit, skip next instruction
  dec ch                    ;decrement cycle count
  xor bh,bh                 ;clear XH
  @@AXXSkip:                ;
  add bx,dx                 ;add absolute to index
  adc ah,0                  ;increment data bank if necessary
  cmp [ePE],0               ;
  je SHORT @@AXXSkip2             ;if in native mode, skip next instruction
  xor ah,ah                 ;clear data bank
  @@AXXSkip2:               ;
  ret                       ;
ENDP
eAXX MACRO
  call AbsoluteIndexX
ENDM
AbsoluteIndexY PROC       ;Set AH:BX to 24-bit effective address, Absolute-Indexed-Y mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get AAL
  add ch,3                  ;increment cycle count +3
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  inc WORD PTR [ePCL]       ;
  inc WORD PTR [ePCL]       ;increment program counter +2
  mov ah,[eDB]              ;get data bank register
  mov bx,WORD PTR [eYL]     ;get Y index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@AXYSkip              ;is 16bit, skip next instruction
  dec ch                    ;decrement cycle count
  xor bh,bh                 ;clear YH
  @@AXYSkip:                ;
  add bx,dx                 ;add absolute to index
  adc ah,0                  ;increment data bank if necessary
  cmp [ePE],0               ;
  je SHORT @@AXYSkip2             ;if in native mode, skip next instruction
  xor ah,ah                 ;clear data bank
  @@AXYSkip2:               ;
  ret
ENDP
eAXY MACRO
  call AbsoluteIndexY
ENDM
AbsoluteLongIndexedX PROC ;Set AH:BX to 24-bit effective address, Absolute-Long-Indexed-X mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get AAL
  add ch,3                  ;increment cycle count +3
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  add WORD PTR [ePCL],3     ;increment program counter +3
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAB
  mov ah,al                 ;preserve AAB
  mov bx,WORD PTR [eXL]     ;get X index register
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@ALXSkip              ;is 16bit, skip next instruction
  xor bh,bh                 ;clear XH
  @@ALXSkip:                ;
  add bx,dx                 ;add absolute to index
  adc ah,0                  ;increment data bank if necessary
  cmp [ePE],0               ;
  je SHORT @@ALXSkip2             ;if in native mode, skip next instruction
  xor ah,ah                 ;clear data bank
  @@ALXSkip2:               ;
  ret                       ;
ENDP
eALX MACRO
  call AbsoluteLongIndexedX
ENDM
DirectIndirect PROC       ;
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  add ch,3                  ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DIDSkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DIDSkip3:               ;
  call MemRead              ;get AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov bh,al                 ;preserve AAH
  mov bl,dl                 ;retrieve AAL
  mov ah,[eDB]              ;get data bank
  ret                       ;
ENDP
eDID MACRO
  call DirectIndirect
ENDM
DirectIndirectLong PROC
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get program counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  add ch,4                  ;increment cycle count
  mov bx,WORD PTR [eDL]     ;get direct register
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset
  cmp [eDL],0               ;
  jz SHORT @@DILSkip3             ;is DL page aligned?
  inc ch                    ;if not, increment cycle count
  @@DILSkip3:               ;
  call MemRead              ;get AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAB
  mov ah,al                 ;set output (AH) to AAB
  mov bx,dx                 ;set output (BX) to AA
  ret                       ;
ENDP
eDIL MACRO
  call DirectIndirectLong
ENDM
StackRelative PROC        ;Set AH:BX to 24-bit effective address, Stack-Relative mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get progam counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  inc ch                    ;increment cycle count
  mov bx,WORD PTR [eSL]     ;get stack pointer
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset into output (BX)
  inc ch                    ;increment cycle count
  ret                       ;
ENDP
eSTR MACRO
  call StackRelative
ENDM
StackRelativeIndirectY PROC   ;Set AH:BX to 24-bit effective address, Stack-Relative mode
  mov ah,[ePB]              ;get program bank
  mov bx,WORD PTR [ePCL]    ;get progam counter
  call MemRead              ;get offset
  inc WORD PTR [ePCL]       ;increment program counter
  mov bx,WORD PTR [eSL]     ;get stack pointer
  xor ah,ah                 ;clear ah
  add bx,ax                 ;add offset into output (BX)
  add ch,5                  ;increment cycle count +5
  call MemRead              ;get AAL
  mov dl,al                 ;preserve AAL
  inc bx                    ;increment internal pointer
  call MemRead              ;get AAH
  mov dh,al                 ;preserve AAH
  mov bx,WORD PTR [eYL]     ;get Y index
  test [ePx],10h             ;test for 8-bit mode
  jz SHORT @@SIISkip              ;is 16bit, skip next instruction
  xor bh,bh                 ;clear YH
  @@SIISkip:                ;
  mov ah,[eDB]              ;get data bank
  add bx,dx                 ;add AA to index
  adc ah,0                  ;increment data bank if necessary
  ret                       ;
ENDP
eSII MACRO
  call StackRelativeIndirectY
ENDM


eRead MACRO    ;AH:BX points to memory location, eP[m] picks size, OUT=DX,c
  LOCAL @@eReadEnd
  call MemRead
  mov dl,al
  inc ch
  test [ePm],20h
  jnz SHORT @@eReadEnd
  inc bx
  call MemRead
  dec bx
  mov dh,al
  inc ch
  stc
  @@eReadEnd:
ENDM
eWrite MACRO   ;AH:BX points to memory location, eP[m] picks size, IN=DX
  LOCAL @@eWriteEnd
  mov al,dl
  call MemWrite
  inc ch
  test [ePm],20h
  jnz SHORT @@eWriteEnd
  inc bx
  mov al,dh
  call MemWrite
  dec bx
  inc ch
  stc
  @@eWriteEnd:
ENDM
exRead MACRO    ;AH:BX points to memory location, eP[x] picks size, OUT=DX,c
  LOCAL @@exReadEnd
  call MemRead
  mov dl,al
  inc ch
  test [ePx],10h
  jnz SHORT @@exReadEnd
  inc bx
  call MemRead
  dec bx
  mov dh,al
  inc ch
  stc
  @@exReadEnd:
ENDM
exWrite MACRO   ;AH:BX points to memory location, eP[x] picks size, IN=DX
  LOCAL @@exWriteEnd
  mov al,dl
  call MemWrite
  inc ch
  test [ePx],10h
  jnz SHORT @@exWriteEnd
  inc bx
  mov al,dh
  call MemWrite
  dec bx
  inc ch
  stc
  @@exWriteEnd:
ENDM

eSETNZ MACRO               ;Set Pn and Pz based on x86 flags
  LOCAL eSETNZa,eSETNZb      ;Define local jump tags
  mov [ePn],0                ;clear negative flag
  mov [ePz],0                ;clear zero flag
  jns SHORT eSETNZa          ;jump if not signed
  mov [ePn],80h              ;  set negative flag
  eSETNZa:                   ;
  jnz SHORT eSETNZb          ;jump if not zero
  mov [ePz],2h               ;  set zero flag
  eSETNZb:                   ;
ENDM
eSETZ MACRO
  LOCAL eSETZS
  mov [ePz],0
  jnz SHORT eSETZS
  mov [ePz],2h
  eSETZS:
ENDM
eGETC MACRO                ;get emulated carry into x86 carry
  LOCAL @@eGETC1             ;define local label
  test [ePc],1               ;get carry bit of eP into zero flag, clear carry
  jz SHORT @@eGETC1                ;if it is zero skip next instruction
  stc                        ;set carry
  @@eGETC1:                  ;
ENDM
eSETC MACRO                ;get x86 carry into emulated carry
  LOCAL @@eSETC1             ;define local labels
  mov [ePc],0                ; clear carry
  jnc SHORT @@eSETC1                ;if it is zero skip next instruction
  mov [ePc],1                ; set carry
  @@eSETC1:                  ;
ENDM
eSETNVZ MACRO
  LOCAL @@SNVZS
  eSETNZ
  mov [ePv],0
  jno @@SNVZS
  mov [ePv],40h
  @@SNVZS:
ENDM
eSETCNZ MACRO
  eSETC
  eSETNZ
ENDM
eCMPV MACRO
  mov [ePv],dl
  and [ePv],40h
  mov [ePn],dl
  mov [ePn],80h
ENDM
eSETMF MACRO          ;Set full math flags
  LOCAL @@SF1           ;Define local jump vars
  eSETCNZ               ;set C,N,Z flags
  mov [ePv],0           ;clear overflow flag
  jno SHORT @@SF1       ;if overflow, skip next instruction
  mov [ePv],40h         ;  set overflow flag
  @@SF1:                ;endif
ENDM
eConFlags MACRO       ;Condense emulated flags into one byte in DL
  mov dl,[ePn]          ;get negative flag
  or dl,[ePv]           ;get overflow flag 
  or dl,[ePm]           ;get accumulator select flag
  or dl,[ePx]           ;get index select flag
  or dl,[ePd]           ;get decimal flag
  or dl,[ePi]           ;get interrupt flag
  or dl,[ePz]           ;get zero flag
  or dl,[ePc]           ;get carry flag
ENDM
eExpFlags MACRO       ;Split eP byte (DL) into multi-bytes (eP[])
  LOCAL eExpFlagsS
  mov [ePn],80h         ;Set mask
  mov [ePv],40h         ;Set mask
  mov [ePm],20h         ;Set mask
  mov [ePx],10h         ;Set mask
  mov [ePd],08h         ;Set mask
  mov [ePi],04h         ;Set mask
  mov [ePz],02h         ;Set mask
  mov [ePc],01h         ;Set mask
  and [ePn],dl          ;Get flag value
  and [ePv],dl          ;Get flag value
  and [ePm],dl          ;Get flag value
  and [ePx],dl          ;Get flag value
  and [ePd],dl          ;Get flag value
  and [ePi],dl          ;Get flag value
  and [ePz],dl          ;Get flag value
  and [ePc],dl          ;Get flag value
  cmp [ePx],0
  je eExpFlagsS         ;In 8-bit mode, clear upper 8-bits of index regs
  mov [eXH],0
  mov [eYH],0
  eExpFlagsS:
ENDM

OpCodeEmu PROC
Op00:  ;BRK
  call eBRK
  jmp EndOp
Op01:  ;ORA (d,x)  Direct Indexed Indirect
  eDXI                     ;Direct Indexed Indirect address set
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op01bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op01bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op02:  ;COP s
  call eCOP
  jmp  EndOp
Op03:  ;ORA d,s
  eSTR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op03bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op03bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op04:  ;TSB d
  eDIR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op04bit16           ;Case 8-bit mode
  test dl,BYTE PTR [eAL]
  eSETZ
  or dl,BYTE PTR [eAL]     ;execute instruction
  eWrite
  jmp EndOp
  Op04bit16:             ;Case 16-bit mode
  test dx,WORD PTR [eAL]
  eSETZ
  or dx,WORD PTR [eAL]     ;execute instruction
  eWrite
  jmp  EndOp
Op05:  ;ORA d
  eDIR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op05bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op05bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op06:
  eDIR
  eRead
  jc @@Op06S
  shl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op06S:
  shl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op07:  ;ORA [d]
  eDIL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op07bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op07bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op08:  ;PHP i
  eConFlags
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,dl
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op08S
  mov bh,1
  Op08S:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op09:  ;ORA #
  eIMM
  jc SHORT Op09bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op09bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op0A:                  ;ASL A
  test [ePm],20h         ;
  jnz @@Op0AS            ;
  shl WORD PTR [eAL],1   ;
  eSETCNZ                ;
  jmp EndOp              ;
  @@Op0AS:               ;
  shl [eAL],1            ;
  eSETCNZ                ;
  jmp  EndOp             ;
Op0B:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eDH]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op0BS
  mov bh,1
  Op0BS:
  mov al,[eDL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op0BS2
  mov bh,1
  Op0BS2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op0C:
  eABS
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op0Cbit16           ;Case 8-bit mode
  test dl,BYTE PTR [eAL]
  eSETZ
  or dl,BYTE PTR [eAL]     ;execute instruction
  eWrite
  jmp EndOp
  Op0Cbit16:             ;Case 16-bit mode
  test dx,WORD PTR [eAL]
  eSETZ
  or dx,WORD PTR [eAL]     ;execute instruction
  eWrite
  jmp  EndOp
Op0D:  ;ORA a
  eABS
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op0Dbit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op0Dbit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op0E:
  eABS
  eRead
  jc @@Op0ES
  shl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op0ES:
  shl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op0F:  ;ORA al
  eABL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op0Fbit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op0Fbit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op10:
  eREL
  cmp [ePn],0
  jne Op10S
  add WORD PTR [ePCL],dx
  Op10S:
  jmp  EndOp
Op11:  ;ORA (d),y
  eDII
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op11bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op11bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op12:  ;ORA (d)
  eDID
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op12bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op12bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op13:  ;ORA (d,s),y
  eSII
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op13bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op13bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op14:
  eDIR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op14bit16           ;Case 8-bit mode
  mov al,BYTE PTR [eAL]
  test dl,al
  eSETZ
  not al
  and dl,al                ;execute instruction
  eWrite
  jmp EndOp
  Op14bit16:             ;Case 16-bit mode
  mov ax,WORD PTR [eAL]
  test dx,ax
  eSETZ                    ;set zero and negative flags
  not ax
  and dx,ax                 ;execute instruction
  eWrite
  jmp  EndOp
Op15:  ;ORA d,x
  eDXX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op15bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op15bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op16:
  eDXX
  eRead
  jc @@Op16S
  shl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op16S:
  shl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op17:  ;ORA [d],y
  eDXL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op17bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op17bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op18:
  mov [ePc],0
  jmp  EndOp
Op19:  ;ORA a,y
  eAXY
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op19bit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op19bit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op1A:        ;inc A
  test [ePm],20h
  jnz SHORT Op1AS
  inc WORD PTR [eAL]
  eSETNZ
  jmp EndOp
  Op1AS:
  inc [eAL]
  eSETNZ
  jmp  EndOp
Op1B: ;TCS
  test [ePm],20h
  jnz SHORT Op1BS
  mov ax,WORD PTR [eAL]
  mov WORD PTR [eSL],ax
  jmp EndOp
  Op1BS:
  mov al,[eAL]
  mov [eSL],al
  jmp  EndOp
Op1C:
  eABS
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op14bit1C           ;Case 8-bit mode
  mov al,BYTE PTR [eAL]
  test dl,al
  eSETZ
  not al
  and dl,al                ;execute instruction
  eWrite
  jmp EndOp
  Op14bit1C:             ;Case 16-bit mode
  mov ax,WORD PTR [eAL]
  test dx,ax
  eSETZ                    ;set zero and negative flags
  not ax
  and dx,ax                 ;execute instruction
  eWrite
  jmp  EndOp
Op1D:  ;ORA a,x
  eAXX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op1Dbit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op1Dbit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op1E:
  eAXX
  eRead
  jc @@Op1ES
  shl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op1ES:
  shl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op1F:  ;ORA al,x
  eALX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op1Fbit16           ;Case 8-bit mode
  or BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op1Fbit16:             ;Case 16-bit mode
  or WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op20:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  inc WORD PTR [ePCL]
  mov ah,0
  mov bx,WORD PTR [eSL]
  mov al,[ePCH]
  call MemWrite
  dec bx
  mov al,[ePCL]
  call MemWrite
  dec WORD PTR [eSL]
  dec WORD PTR [eSL]
  cmp [ePE],0
  je SHORT Op20S
  mov [eSH],1
  Op20S:
  mov WORD PTR [ePCL],dx
  jmp  EndOp
Op21:   ;AND (d,x)
  eDXI
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op21bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op21bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op22:
  mov bx,WORD PTR [eSL]
  mov ah,0
  add WORD PTR [ePCL],2
  mov al,[ePB]
  call MemWrite
  dec bx
  mov al,[ePCH]
  call MemWrite
  dec bx
  mov al,[ePCL]
  call MemWrite
  dec bx
  mov WORD PTR [eSL],bx

  mov bx,WORD PTR [ePCL]
  sub bx,2

  mov ah,[ePB]
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc bx
  call MemRead
  mov [ePB],al
  jmp  EndOp
Op23:   ;AND d,s
  eSTR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op23bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op23bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op24:
  eDIR
  eRead
  jc SHORT Op24S
  test [eAL],dl
  eSETZ
  eCMPV
  jmp EndOp
  Op24S:
  test WORD PTR [eAL],dx
  eSETZ
  eCMPV
  jmp  EndOp
Op25:   ;AND d
  eDIR
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op25bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op25bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op26:
  eDIR
  eRead
  jc @@Op26S
  eGETC
  rcl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op26S:
  eGETC
  rcl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op27:   ;AND [d]
  eDIL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op27bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op27bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op28:
  ePOP
  mov dl,al
  eExpFlags
  jmp  EndOp
Op29:   ;AND #
  eIMM
  jc SHORT Op29bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op29bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op2A:
  test [ePm],20h         ;
  jnz @@Op2AS            ;
  eGETC
  rcl WORD PTR [eAL],1   ;
  eSETCNZ                ;
  jmp EndOp              ;
  @@Op2AS:               ;
  eGETC
  rcl [eAL],1            ;
  eSETCNZ                ;
  jmp  EndOp
Op2B:
  ePOP
  mov dl,al
  ePOP
  mov dh,al
  mov WORD PTR [eDL],dx
  or dx,0
  eSETNZ
  jmp  EndOp
Op2C:
  eABS
  eRead
  jc SHORT Op2CS
  test [eAL],dl
  eSETZ
  eCMPV
  jmp EndOp
  Op2CS:
  test WORD PTR [eAL],dx
  eSETZ
  eCMPV
  jmp  EndOp
Op2D:   ;AND a
  eABS
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op2Dbit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op2Dbit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op2E:
  eABS
  eRead
  jc @@Op2ES
  eGETC
  rcl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op2ES:
  eGETC
  rcl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op2F:   ;AND al
  eABL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op2Fbit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op2Fbit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op30:
  eREL
  cmp [ePn],0
  je SHORT Op30S
  add WORD PTR [ePCL],dx
  Op30S:
  jmp  EndOp
Op31:   ;AND (d),y
  eDII
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op31bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op31bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op32:   ;AND (d)
  eDID
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op32bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op32bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op33:   ;AND (d,s),y
  eSII
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op33bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op33bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op34:
  eDXX
  eRead
  jc SHORT Op34S
  test [eAL],dl
  eSETZ
  eCMPV
  jmp EndOp
  Op34S:
  test WORD PTR [eAL],dx
  eSETZ
  eCMPV
  jmp  EndOp
Op35:   ;AND d,x
  eDXX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op35bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op35bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op36:
  eDXX
  eRead
  jc @@Op36S
  eGETC
  rcl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op36S:
  eGETC
  rcl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op37:   ;AND [d],y
  eDXL
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op37bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op37bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op38:
  mov [ePc],1
  jmp  EndOp
Op39:   ;AND AND a,y
  eAXY
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op39bit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op39bit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op3A:
  test [ePm],20h
  jnz SHORT Op3AS
  dec WORD PTR [eAL]
  eSETNZ
  jmp EndOp
  Op3AS:
  dec [eAL]
  eSETNZ
  jmp  EndOp
Op3B:
  mov ax,WORD PTR [eSL]
  mov WORD PTR [eAL],ax
  or ax,0
  eSETNZ
  jmp EndOp
Op3C:
  eAXX
  eRead
  jc SHORT Op3CS
  test [eAL],dl
  eSETZ
  eCMPV
  jmp EndOp
  Op3CS:
  test WORD PTR [eAL],dx
  eSETZ
  eCMPV
  jmp  EndOp
Op3D:   ;AND a,x
  eAXX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op3Dbit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op3Dbit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op3E:
  eAXX
  eRead
  jc @@Op3ES
  eGETC
  rcl dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op3ES:
  eGETC
  rcl dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op3F:   ;AND al,x
  eALX
  eRead                    ;Get Byte/Word (carry set=word)
  jc SHORT Op3Fbit16           ;Case 8-bit mode
  and BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op3Fbit16:             ;Case 16-bit mode
  and WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op40:
  mov bx,WORD PTR [eSL]
  mov ah,0
  cmp [ePE],0
  je SHORT Op40S
  inc bl
  call MemRead
  mov dl,al
  eExpFlags
  inc bl
  call MemRead
  mov [ePCL],al
  inc bl
  call MemRead
  mov [ePCH],al
  add [eSL],3
  jmp EndOp
  Op40S:
  inc bx
  call MemRead
  mov dl,al
  eExpFlags
  inc bx
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc bx
  call MemRead
  mov [ePB],al
  add [eSL],4
  jmp  EndOp
Op41:   ;EOR (d,x)
  eDXI
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op41bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op41bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op42:
  jmp  EndOp
Op43:   ;EOR d,s
  eSTR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op43bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op43bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op44:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dh,al
  inc bx
  call MemRead
  mov dl,al
  ror ecx,16
  mov cx,WORD PTR [eAL]
  inc cx
  Op44Loop:
  mov bx,WORD PTR [eXL]
  mov ah,dl
  call MemRead
  mov ah,dh
  mov bx,WORD PTR [eYL]
  call MemWrite
  dec WORD PTR [eXL]
  dec WORD PTR [eYL]
  loop Op44Loop
  ror ecx,16
  mov WORD PTR [eAL],0ffffh
  mov BYTE PTR [eDB],ah
  jmp  EndOp
Op45:   ;EOR d
  eDIR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op45bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op45bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op46:
  eDIR
  eRead
  jc @@Op46S
  shr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op46S:
  shr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op47:   ;EOR [d],y
  eDIL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op47bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op47bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op48:
  test [ePm],20h
  jnz SHORT Op48S3
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eAH]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op48S
  mov bh,1
  Op48S:
  mov al,[eAL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op48S2
  mov bh,1
  Op48S2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
  Op48S3:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eAL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op48S4
  mov bh,1
  Op48S4:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op49:   ;EOR #
  eIMM
  jc SHORT Op49bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op49bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op4A:
  test [ePm],20h         ;
  jnz @@Op4AS            ;
  shl WORD PTR [eAL],1   ;
  eSETCNZ                ;
  jmp EndOp              ;
  @@Op4AS:               ;
  shl [eAL],1            ;
  eSETCNZ                ;
  jmp  EndOp
Op4B:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[ePB]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op4BS
  mov bh,1
  Op4BS:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op4C:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  jmp  EndOp
Op4D:   ;EOR a
  eABS
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op4Dbit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op4Dbit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op4E:
  eABS
  eRead
  jc @@Op4ES
  shr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op4ES:
  shr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op4F:   ;EOR al
  eABL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op4Fbit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op4Fbit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op50:
  eREL
  cmp [ePv],0
  jne SHORT Op50S
  add WORD PTR [ePCL],dx
  Op50S:
  jmp  EndOp
Op51:   ;EOR (d),y
  eDII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op51bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op51bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op52:   ;EOR (d)
  eDID
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op52bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op52bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op53:   ;EOR (d,s),y
  eSII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op53bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op53bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op54:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dh,al
  inc bx
  call MemRead
  mov dl,al
  ror ecx,16
  mov cx,WORD PTR [eAL]
  inc cx
  Op54Loop:
  mov bx,WORD PTR [eXL]
  mov ah,dl
  call MemRead
  mov ah,dh
  mov bx,WORD PTR [eYL]
  call MemWrite
  inc WORD PTR [eXL]
  inc WORD PTR [eYL]
  loop Op54Loop
  ror ecx,16
  mov WORD PTR [eAL],0ffffh
  mov [eDB],dh
  jmp  EndOp
Op55:   ;EOR d,x
  eDXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op55bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op55bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op56:
  eDXX
  eRead
  jc @@Op56S
  shr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op56S:
  shr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op57:   ;EOR [d],y
  eDXL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op57bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op57bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op58:
  mov [ePi],0
  jmp  EndOp
Op59:   ;EOR a,y
  eAXY
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op59bit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op59bit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op5A:
  test [ePx],10h
  jnz SHORT Op5AS3
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eYH]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op5AS
  mov bh,1
  Op5AS:
  mov al,[eYL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op5AS2
  mov bh,1
  Op5AS2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
  Op5AS3:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eYL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op5AS4
  mov bh,1
  Op5AS4:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op5B:
  mov ax,WORD PTR [eAL]
  mov WORD PTR[eDL],ax
  or ax,ax
  eSETNZ
  jmp EndOp
Op5C:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc bx
  call MemRead
  mov [ePB],al
  jmp  EndOp
Op5D:   ;EOR a,x
  eAXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op5Dbit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op5Dbit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op5E:
  eAXX
  eRead
  jc @@Op5ES
  shr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op5ES:
  shr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op5F:   ;EOR al,x
  eALX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op5Fbit16           ;Case 8-bit mode
  xor BYTE PTR [eAL],dl     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp EndOp                
  Op5Fbit16:             ;Case 16-bit mode
  xor WORD PTR [eAL],dx     ;execute instruction
  eSETNZ                    ;set zero and negative flags
  jmp  EndOp
Op60:
  cmp [ePE],0
  je SHORT Op60S
  mov bx,WORD PTR [eSL]
  mov ah,0
  inc bl
  call MemRead
  mov [ePCL],al
  inc bl
  call MemRead
  mov [ePCH],al
  inc WORD PTR [ePCL]
  add [eSL],2
  jmp EndOp
  Op60S:
  mov bx,WORD PTR [eSL]
  mov ah,0
  inc bx
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc WORD PTR [ePCL]
  add [eSL],2
  jmp  EndOp
Op61:
  eDXI
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op61bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op61bitF      
  Op61bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op61bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op62:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  inc WORD PTR [ePCL]
  inc WORD PTR [ePCL]
  add dx,WORD PTR [ePCL]
  xor ah,ah
  mov bx,WORD PTR [eSL]
  mov al,dh
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op62S1
  mov bh,1
  Op62S1:
  mov al,dl
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op62S2
  mov bh,1
  Op62S2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op63:
  eSTR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op63bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op63bitF      
  Op63bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op63bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op64:
  eDIR
  mov dx,0
  eWRITE
  jmp  EndOp
Op65:
  eDIR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op65bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op65bitF      
  Op65bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op65bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op66:
  eDIR
  eRead
  jc @@Op66S
  eGETC
  rcr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op66S:
  eGETC
  rcr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op67:
  eDIL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op67bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op67bitF      
  Op67bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op67bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op68:
  test [ePm],20h
  jz SHORT Op68S
  ePOP
  mov [eAL],al
  or al,0
  eSETNZ
  jmp EndOp
  Op68S:
  ePOP
  mov dl,al
  ePOP
  mov dh,al
  mov WORD PTR [eAL],dx
  or dx,0
  eSETNZ
  jmp  EndOp
Op69:
  eIMM
  jc SHORT Op69bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op69bitF      
  Op69bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op69bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op6A:
  test [ePm],20h         ;
  jnz @@Op6AS            ;
  eGETC
  rcl WORD PTR [eAL],1   ;
  eSETCNZ                ;
  jmp EndOp              ;
  @@Op6AS:               ;
  eGETC
  rcl [eAL],1            ;
  eSETCNZ                ;
  jmp  EndOp
Op6B:   ;RTL i
  mov bx,WORD PTR [eSL]
  mov ah,0
  inc bx
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc bx
  call MemRead
  mov [ePB],al
  add WORD PTR [eSL],3
  inc WORD PTR [ePCL]
  jmp  EndOp
Op6C:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  mov ah,0
  mov bx,dx
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  jmp  EndOp
Op6D:
  eABS
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op6Dbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op6DbitF      
  Op6Dbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op6DbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op6E:
  eABS
  eRead
  jc @@Op6ES
  eGETC
  rcr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op6ES:
  eGETC
  rcr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op6F:
  eABL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op6Fbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op6FbitF      
  Op6Fbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op6FbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op70:
  eREL
  cmp [ePv],0
  je SHORT Op70S
  add WORD PTR [ePCL],dx
  Op70S:
  jmp  EndOp
Op71:
  eDII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op71bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op71bitF      
  Op71bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op71bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op72:
  eDID
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op72bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op72bitF      
  Op72bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op72bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op73:
  eSII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op73bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op73bitF      
  Op73bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op73bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op74:
  eDXX
  mov dx,0
  eWRITE
  jmp  EndOp
Op75:
  eDXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op75bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op75bitF      
  Op75bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op75bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op76:
  eDXX
  eRead
  jc @@Op76S
  eGETC
  rcr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op76S:
  eGETC
  rcr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op77:
  eDXL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op77bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op77bitF      
  Op77bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op77bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op78:
  mov [ePi],4
  jmp  EndOp
Op79:
  eAXY
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op79bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op79bitF      
  Op79bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op79bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op7A:
  test [ePx],10h
  jz SHORT Op7AS
  ePOP
  mov [eYL],al
  or al,0
  eSETNZ
  jmp EndOp
  Op7AS:
  ePOP
  mov dl,al
  ePOP
  mov dh,al
  mov WORD PTR [eYL],dx
  or dx,0
  eSETNZ
  jmp  EndOp
Op7B:
  mov ax,WORD PTR [eDL]
  mov WORD PTR [eAL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
Op7C:
  eAXX
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  jmp  EndOp
Op7D:
  eAXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op7Dbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op7DbitF      
  Op7Dbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op7DbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op7E:
  eAXX
  eRead
  jc @@Op7ES
  eGETC
  rcr dl,1
  eSETCNZ
  eWrite
  jmp EndOp
  @@Op7ES:
  eGETC
  rcr dx,1
  eSETCNZ
  eWrite
  jmp  EndOp
Op7F:
  eALX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT Op7Fbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc BYTE PTR [eAL],dl     ;execute instruction
  jmp SHORT Op7FbitF      
  Op7Fbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  adc WORD PTR [eAL],dx     ;execute instruction
  Op7FbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
Op80:
  eREL
  add WORD PTR [ePCL],dx
  jmp  EndOp
Op81:
  eDXI
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op82:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  inc WORD PTR [ePCL]
  inc WORD PTR [ePCL]
  add WORD PTR [ePCL],dx
  jmp  EndOp
Op83:
  eSTR
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op84:
  eDIR
  mov dx,WORD PTR [eYL]
  exWrite
  jmp  EndOp
Op85:
  eDIR
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op86:
  eDIR
  mov dx,WORD PTR [eXL]
  exWrite
  jmp  EndOp
Op87:
  eDIL
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op88:
  test [ePx],10h
  jz SHORT Op88S
  dec [eYL]
  eSETNZ
  jmp EndOp
  Op88S:
  dec WORD PTR [eYL]
  eSETNZ
  jmp  EndOp
Op89:
  eIMM
  eRead
  jc SHORT Op89S
  test [eAL],dl
  eSETNZ
  jmp EndOp
  Op89S:
  test WORD PTR [eAL],dx
  eSETNZ
  jmp  EndOp
Op8A:
  mov ax,WORD PTR[eXL]
  mov WORD PTR[eAL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
Op8B:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eDB]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT Op8BS
  mov bh,1
  Op8BS:
  mov WORD PTR [eSL],bx
  jmp  EndOp
Op8C:
  eABS
  mov dx,WORD PTR [eYL]
  exWrite
  jmp  EndOp
Op8D:
  eABS
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op8E:
  eABS
  mov dx,WORD PTR [eXL]
  exWrite
  jmp  EndOp
Op8F:
  eABL
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op90:
  eREL
  cmp [ePc],0
  jne SHORT Op90S
  add WORD PTR [ePCL],dx
  Op90S:
  jmp  EndOp
Op91:
  eDII
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op92:
  eDID
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op93:
  eSII
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op94:
  eDXX
  mov dx,WORD PTR [eYL]
  exWrite
  jmp  EndOp
Op95:
  eDXX
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op96:
  eDXY
  mov dx,WORD PTR [eXL]
  exWrite
  jmp  EndOp
Op97:
  eDXL
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op98:
  mov ax,WORD PTR[eYL]
  mov WORD PTR[eAL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
Op99:
  eAXY
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op9A:
  mov ax,WORD PTR [eXL]
  cmp [ePE],0
  je SHORT Op9AS
  mov ah,1
  Op9AS:
  mov WORD PTR [eSL],ax
  jmp  EndOp
Op9B:
  mov ax,WORD PTR [eXL]
  mov WORD PTR [eYL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
Op9C:
  eABS
  mov dx,0
  eWrite
  jmp  EndOp
Op9D:
  eAXX
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
Op9E:
  eAXX
  mov dx,0
  eWrite
  jmp  EndOp
Op9F:
  eALX
  mov dx,WORD PTR [eAL]
  eWrite
  jmp  EndOp
OpA0:
  eIMMX
  jc SHORT OpA0S
  mov [eYL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpA0S:
  mov WORD PTR [eYL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA1:
  eDXI
  eRead
  jc SHORT @@OpA1S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpA1S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp EndOp
OpA2:
  eIMMX
  jc SHORT OpA2S
  mov [eXL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpA2S:
  mov WORD PTR [eXL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA3:
  eSTR
  eRead
  jc SHORT @@OpA3S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpA3S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA4:
  eDIR
  exRead
  jc SHORT OpA4S
  mov [eYL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpA4S:
  mov WORD PTR [eYL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA5:
  eDIR
  eRead
  jc SHORT @@OpA5S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpA5S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA6:
  eDIR
  exRead
  jc SHORT OpA6S
  mov [eXL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpA6S:
  mov WORD PTR [eXL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA7:
  eDIL
  eRead
  jc SHORT @@OpA7S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpA7S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpA8:
  test [ePx],10h
  jnz SHORT OpA8S
  mov ax,WORD PTR [eAL]
  mov WORD PTR [eYL],ax
  or ax,ax
  eSETNZ
  jmp EndOp
  OpA8S:
  mov al,[eAL]
  mov [eYL],al
  or al,al
  eSETNZ
  jmp  EndOp
OpA9:
  eIMM
  jc SHORT @@OpA9S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpA9S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpAA:
  test [ePx],10h
  jnz SHORT OpAAS
  mov ax,WORD PTR [eAL]
  mov WORD PTR [eXL],ax
  or ax,ax
  eSETNZ
  jmp EndOp
  OpAAS:
  mov al,[eAL]
  mov [eXL],al
  or al,al
  eSETNZ
  jmp  EndOp
OpAB:
  ePOP
  mov [eDB],al
  or al,0
  eSETNZ
  jmp  EndOp
OpAC:
  eABS
  exRead
  jc SHORT OpACS
  mov [eYL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpACS:
  mov WORD PTR [eYL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpAD:
  eABS
  eRead
  jc SHORT @@OpADS
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpADS:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpAE:
  eABS
  exRead
  jc SHORT OpAES
  mov [eXL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpAES:
  mov WORD PTR [eXL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpAF:
  eABL
  eRead
  jc SHORT @@OpAFS
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpAFS:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB0:
  eREL
  cmp [ePc],0
  je SHORT OpB0S
  add WORD PTR [ePCL],dx
  OpB0S:
  jmp  EndOp
OpB1:
  eDII
  eRead
  jc SHORT @@OpB1S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB1S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB2:
  eDID
  eRead
  jc SHORT @@OpB2S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB2S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB3:
  eSII
  eRead
  jc SHORT @@OpB3S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB3S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB4:
  eDXX
  exRead
  jc SHORT OpB4S
  mov [eYL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpB4S:
  mov WORD PTR [eYL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB5:
  eDXX
  eRead
  jc SHORT @@OpB5S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB5S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB6:
  eDXY
  exRead
  jc SHORT OpB6S
  mov [eXL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpB6S:
  mov WORD PTR [eXL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB7:
  eDXL
  eRead
  jc SHORT @@OpB7S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB7S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpB8:
  mov [ePv],0
  jmp  EndOp
OpB9:
  eAXY
  eRead
  jc SHORT @@OpB9S
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpB9S:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpBA:
  test [ePx],10h
  jz SHORT OpBAS
  mov al,[eSL]
  mov [eXL],al
  or al,al
  eSETNZ
  jmp EndOp
  OpBAS:
  mov ax,WORD PTR [eSL]
  mov WORD PTR [eXL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
OpBB:
  test [ePx],10h
  jz SHORT OpBBS
  mov al,[eYL]
  mov [eXL],al
  or al,al
  eSETNZ
  jmp EndOp
  OpBBS:
  mov ax,WORD PTR [eYL]
  mov WORD PTR [eXL],ax
  or ax,ax
  eSETNZ
  jmp  EndOp
OpBC:
  eAXX
  exRead
  jc SHORT OpBCS
  mov [eYL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpBCS:
  mov WORD PTR [eYL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpBD:
  eAXX
  eRead
  jc SHORT @@OpBDS
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpBDS:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpBE:
  eAXY
  exRead
  jc SHORT OpBES
  mov [eXL],dl
  or dl,dl
  eSETNZ
  jmp EndOp
  OpBES:
  mov WORD PTR [eXL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpBF:
  eALX
  eRead
  jc SHORT @@OpBFS
  mov [eAL],dl
  or dl,dl
  eSETNZ
  jmp  EndOp
  @@OpBFS:
  mov WORD PTR [eAL],dx
  or dx,dx
  eSETNZ
  jmp  EndOp
OpC0:
  eIMMX
  jc SHORT OpC0S
  cmp [eYL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC0S:
  cmp WORD PTR [eYL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC1:     ;CMP
  eDXI
  eRead
  jc SHORT OpC1S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC1S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC2:
  eIMM8
  not dl
  mov dh,dl
  eConFlags
  and dl,dh
  eExpFlags
  jmp  EndOp
OpC3:
  eSTR
  eRead
  jc SHORT OpC3S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC3S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC4:
  eDIR
  exRead
  jc SHORT OpC4S
  cmp [eYL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC4S:
  cmp WORD PTR [eYL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC5:
  eDIR
  eRead
  jc SHORT OpC5S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC5S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC6:
  eDIR
  eRead
  jc SHORT OpC6S
  dec dl
  eSETNZ
  jmp SHORT OpC6S2
  OpC6S:
  dec dx
  eSETNZ
  OpC6S2:
  eWrite
  jmp  EndOp
OpC7:
  eDIL
  eRead
  jc SHORT OpC7S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC7S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpC8:              ;Inc Y
  test [ePx],10h
  jnz SHORT OpC8S
  inc WORD PTR [eYL]
  eSETNZ
  jmp EndOp
  OpC8S:
  inc [eYL]
  eSETNZ
  jmp  EndOp
OpC9:
  eIMM
  jc SHORT OpC9S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpC9S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpCA:
  test [ePx],10h
  jnz SHORT OpCAS
  dec WORD PTR [eXL]
  eSETNZ
  jmp EndOp
  OpCAS:
  dec [eXL]
  eSETNZ
  jmp  EndOp
OpCB:  ;WAIT


  jmp  EndOp
OpCC:
  eABS
  exRead
  jc SHORT OpCCS
  cmp [eYL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpCCS:
  cmp WORD PTR [eYL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpCD:  ;CMP a
  eABS
  eRead
  jc SHORT OpCDS
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpCDS:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpCE:
  eABS
  eRead
  jc SHORT OpCES
  dec dl
  eSETNZ
  jmp SHORT OpCES2
  OpCES:
  dec dx
  eSETNZ
  OpCES2:
  eWrite
  jmp  EndOp
OpCF:
  eABL
  eRead
  jc SHORT OpCFS
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpCFS:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD0:
  eREL
  cmp [ePz],0
  jne SHORT OpD0S
  cmp [eSS1],0
  jne OpD0S
  add WORD PTR [ePCL],dx
  OpD0S:
  jmp  EndOp
OpD1:
  eDII
  eRead
  jc SHORT OpD1S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD1S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD2:
  eDID
  eRead
  jc SHORT OpD2S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD2S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD3:
  eSII
  eRead
  jc SHORT OpD3S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD3S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD4:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc WORD PTR [ePCL]
  xor dh,dh
  add dx,WORD PTR [eDL]
  mov bx,dx
  mov ah,[eDB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  xor ah,ah
  mov bx,WORD PTR [eSL]
  mov al,dh
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpD4S1
  mov bh,1
  OpD4S1:
  mov al,dl
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpD4S2
  mov bh,1
  OpD4S2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
OpD5:
  eDXX
  eRead
  jc SHORT OpD5S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD5S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD6:
  eDXX
  eRead
  jc SHORT OpD6S
  dec dl
  eSETNZ
  jmp SHORT OpD6S2
  OpD6S:
  dec dx
  eSETNZ
  OpD6S2:
  eWrite
  jmp  EndOp
OpD7:
  eDXL
  eRead
  jc SHORT OpD7S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD7S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpD8:
  mov [ePd],0
  jmp  EndOp
OpD9:
  eAXY
  eRead
  jc SHORT OpD9S
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpD9S:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpDA:
  test [ePx],10h
  jnz SHORT OpDAS3
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eXH]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpDAS
  mov bh,1
  OpDAS:
  mov al,[eXL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpDAS2
  mov bh,1
  OpDAS2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
  OpDAS3:
  mov bx,WORD PTR [eSL]
  mov ah,0
  mov al,[eXL]
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpDAS4
  mov bh,1
  OpDAS4:
  mov WORD PTR [eSL],bx
  jmp  EndOp
OpDB:
  mov [eSTAT],1
  jmp  EndOp
OpDC:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov bh,al
  xor ah,ah
  mov bl,dl
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  inc bx
  call MemRead
  mov [ePB],al
  jmp  EndOp
OpDD:
  eAXX
  eRead
  jc SHORT OpDDS
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpDDS:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpDE:
  eAXX
  eRead
  jc SHORT OpDES
  dec dl
  eSETNZ
  jmp SHORT OpDES2
  OpDES:
  dec dx
  eSETNZ
  OpDES2:
  eWrite
  jmp  EndOp
OpDF:
  eALX
  eRead
  jc SHORT OpDFS
  cmp [eAL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpDFS:
  cmp WORD PTR [eAL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpE0:
  eIMMX
  jc SHORT OpE0S
  cmp [eXL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpE0S:
  cmp WORD PTR [eXL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpE1:
  eDXI
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpE1bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpE1bitF
  OpE1bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpE1bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpE2:
  eIMM8
  mov dh,dl
  eConFlags
  or dl,dh
  eExpFlags
  jmp  EndOp
OpE3:
  eSTR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpE3bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpE3bitF
  OpE3bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpE3bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpE4:
  eDIR
  exRead
  jc SHORT OpE4S
  cmp [eXL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpE4S:
  cmp WORD PTR [eXL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpE5:
  eDIR
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpE5bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpE5bitF
  OpE5bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpE5bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpE6:
  eDIR
  eRead
  jc SHORT OpE6S
  inc dl
  eSETNZ
  jmp SHORT OpE6S2
  OpE6S:
  inc dx
  eSETNZ
  OpE6S2:
  eWrite
  jmp  EndOp
OpE7:
  eDIL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpE7bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpE7bitF
  OpE7bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpE7bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpE8:                    ;Inc X
  test [ePx],10h
  jnz SHORT OpE8S
  inc WORD PTR [eXL]
  eSETNZ
  jmp EndOp
  OpE8S:
  inc [eXL]
  eSETNZ
  jmp  EndOp
OpE9:
  eIMM
  jc SHORT OpE9bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpE9bitF
  OpE9bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpE9bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpEA:  ;NOP
  jmp  EndOp
OpEB:
  mov al,[eAL]
  mov ah,[eAH]
  mov [eAH],al
  mov [eAL],ah
  or ah,ah
  eSETNZ
  jmp  EndOp
OpEC:
  eABS
  exRead
  jc SHORT OpECS
  cmp [eXL],dl
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp EndOp
  OpECS:
  cmp WORD PTR [eXL],dx
  cmc                       ;Complement Carry
  eSETNZ
  eSETC
  jmp  EndOp
OpED:
  eABS
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpEDbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpEDbitF
  OpEDbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpEDbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpEE:
  eABS
  eRead
  jc SHORT OpEES
  inc dl
  eSETNZ
  jmp SHORT OpEES2
  OpEES:
  inc dx
  eSETNZ
  OpEES2:
  eWrite
  jmp  EndOp
OpEF:
  eABL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpEFbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpEFbitF
  OpEFbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpEFbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF0:
  eREL
  cmp [ePz],0
  je SHORT OpF0S
  add WORD PTR [ePCL],dx
  OpF0S:
  jmp  EndOp
OpF1:
  eDII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF1bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF1bitF
  OpF1bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF1bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF2:
  eDID
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF2bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF2bitF
  OpF2bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF2bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF3:
  eSII
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF3bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF3bitF
  OpF3bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF3bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF4:
  mov bx,WORD PTR [ePCL]
  mov ah,[ePB]
  call MemRead
  mov dl,al
  inc bx
  call MemRead
  mov dh,al
  inc WORD PTR [ePCL]
  inc WORD PTR [ePCL]
  xor ah,ah
  mov bx,WORD PTR [eSL]
  mov al,dh
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpF4S1
  mov bh,1
  OpF4S1:
  mov al,dl
  call MemWrite
  dec bx
  cmp [ePE],0
  je SHORT OpF4S2
  mov bh,1
  OpF4S2:
  mov WORD PTR [eSL],bx
  jmp  EndOp
OpF5:
  eDXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF5bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF5bitF
  OpF5bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF5bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF6:
  eDXX
  eRead
  jc SHORT OpF6S
  inc dl
  eSETNZ
  jmp SHORT OpF6S2
  OpF6S:
  inc dx
  eSETNZ
  OpF6S2:
  eWrite
  jmp  EndOp
OpF7:
  eDXL
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF7bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF7bitF
  OpF7bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF7bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpF8:
  mov [ePd],8
  jmp  EndOp
OpF9:
  eAXY
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpF9bit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpF9bitF
  OpF9bit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpF9bitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpFA:
  test [ePx],10h
  jz SHORT OpFAS
  ePOP
  mov [eXL],al
  or al,0
  eSETNZ
  jmp EndOp
  OpFAS:
  ePOP
  mov dl,al
  ePOP
  mov dh,al
  mov WORD PTR [eXL],dx
  or dx,0
  eSETNZ
  jmp  EndOp
OpFB:
  test [ePc],1
  jz SHORT OpFBS
  mov al,[ePE]
  mov [ePc],al
  mov [ePE],1
  jmp EndOp
  OpFBS:
  mov al,[ePE]
  mov [ePc],al
  mov [ePE],0
  jmp  EndOp
OpFC:
  mov bx,WORD PTR [eSL]
  mov ah,0
  inc WORD PTR [ePCL]
  mov al,[ePCH]
  call MemWrite
  dec bx
  mov al,[ePCL]
  call MemWrite
  dec bx
  mov WORD PTR [eSL],bx
  dec WORD PTR [ePCL]
  eAXX
  call MemRead
  mov [ePCL],al
  inc bx
  call MemRead
  mov [ePCH],al
  jmp  EndOp
OpFD:
  eAXX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpFDbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpFDbitF
  OpFDbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpFDbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpFE:
  eAXX
  eRead
  jc SHORT OpFES
  inc dl
  eSETNZ
  jmp SHORT OpFES2
  OpFES:
  inc dx
  eSETNZ
  OpFES2:
  eWrite
  jmp  EndOp
OpFF:
  eALX
  eRead                  ;Get Byte/Word (carry set=word)
  jc SHORT OpFFbit16           ;Case 8-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb BYTE PTR [eAL],dl     ;execute instruction
  cmc                       ;Complement Carry
  jmp SHORT OpFFbitF
  OpFFbit16:             ;Case 16-bit mode
  eGETC                     ;Get Emulated carry flag into x86 carry
  cmc                       ;Complement Carry
  sbb WORD PTR [eAL],dx     ;execute instruction
  cmc                       ;Complement Carry
  OpFFbitF:
  eSETMF                    ;set zero and negative flags
  jmp  EndOp
OpCodeEmu ENDP

JumpTable dw Op00, Op01, Op02, Op03, Op04, Op05, Op06, Op07
          dw Op08, Op09, Op0A, Op0B, Op0C, Op0D, Op0E, Op0F
          dw Op10, Op11, Op12, Op13, Op14, Op15, Op16, Op17
          dw Op18, Op19, Op1A, Op1B, Op1C, Op1D, Op1E, Op1F
          dw Op20, Op21, Op22, Op23, Op24, Op25, Op26, Op27
          dw Op28, Op29, Op2A, Op2B, Op2C, Op2D, Op2E, Op2F
          dw Op30, Op31, Op32, Op33, Op34, Op35, Op36, Op37
          dw Op38, Op39, Op3A, Op3B, Op3C, Op3D, Op3E, Op3F
          dw Op40, Op41, Op42, Op43, Op44, Op45, Op46, Op47
          dw Op48, Op49, Op4A, Op4B, Op4C, Op4D, Op4E, Op4F
          dw Op50, Op51, Op52, Op53, Op54, Op55, Op56, Op57
          dw Op58, Op59, Op5A, Op5B, Op5C, Op5D, Op5E, Op5F
          dw Op60, Op61, Op62, Op63, Op64, Op65, Op66, Op67
          dw Op68, Op69, Op6A, Op6B, Op6C, Op6D, Op6E, Op6F
          dw Op70, Op71, Op72, Op73, Op74, Op75, Op76, Op77
          dw Op78, Op79, Op7A, Op7B, Op7C, Op7D, Op7E, Op7F
          dw Op80, Op81, Op82, Op83, Op84, Op85, Op86, Op87
          dw Op88, Op89, Op8A, Op8B, Op8C, Op8D, Op8E, Op8F
          dw Op90, Op91, Op92, Op93, Op94, Op95, Op96, Op97
          dw Op98, Op99, Op9A, Op9B, Op9C, Op9D, Op9E, Op9F
          dw OpA0, OpA1, OpA2, OpA3, OpA4, OpA5, OpA6, OpA7
          dw OpA8, OpA9, OpAA, OpAB, OpAC, OpAD, OpAE, OpAF
          dw OpB0, OpB1, OpB2, OpB3, OpB4, OpB5, OpB6, OpB7
          dw OpB8, OpB9, OpBA, OpBB, OpBC, OpBD, OpBE, OpBF
          dw OpC0, OpC1, OpC2, OpC3, OpC4, OpC5, OpC6, OpC7
          dw OpC8, OpC9, OpCA, OpCB, OpCC, OpCD, OpCE, OpCF
          dw OpD0, OpD1, OpD2, OpD3, OpD4, OpD5, OpD6, OpD7
          dw OpD8, OpD9, OpDA, OpDB, OpDC, OpDD, OpDE, OpDF
          dw OpE0, OpE1, OpE2, OpE3, OpE4, OpE5, OpE6, OpE7
          dw OpE8, OpE9, OpEA, OpEB, OpEC, OpED, OpEE, OpEF
          dw OpF0, OpF1, OpF2, OpF3, OpF4, OpF5, OpF6, OpF7
          dw OpF8, OpF9, OpFA, OpFB, OpFC, OpFD, OpFE, OpFF

OpEmu PROC NEAR     ;Execute a 65816 instruction
;IN: bx=opcode  OUT: carry,al=EmuStatus
  IFDEF EDEBUG
    call DebugEOL
    call DebugOpCode
  ENDIF
  shl bx,1
  add bx,offset JumpTable
  jmp CS:[bx]
EndOp: 
  ret
OpEmu ENDP

ExecCPU PROC NEAR   ;Single-Step Emulated 65816
  mov ah,[ePB]
  mov bx,WORD PTR [ePCL]
  call MemRead
  mov bl,al
  xor bh,bh
  inc WORD PTR [ePCL]
  call OpEmu
  IFDEF EDEBUG
    call DebugStat
    call DebugWait
  ENDIF
  ret
ExecCPU ENDP

eNMI PROC NEAR
  mov al,[ePB]
  PushAL
  mov al,[ePCH]
  PushAL
  mov al,[ePCL]
  PushAL
  eConFlags              ;
  mov al,dl              ;
  cmp [ePE],1            ;
  jne SHORT @@NMIS       ;
  and al,11101111b       ;
@@NMIS:                     ;
  PushAL                 ;Push flags
  mov ah,0               ;
  mov bx,0fffah          ;
  cmp [ePE],1
  je NMIS2
  mov bx,0ffeah
  NMIS2:
  call MemRead           ;Get Reset Vector
  mov [ePB],0            ;clear program bank
  mov [ePCL],al          ;get reset vector AAL
  inc bx                 ;
  call MemRead           ;
  mov [ePCH],al          ;get reset vector AAH
  ret
ENDP

eIRQ PROC NEAR
  mov al,[ePB]
  PushAL
  mov al,[ePCH]
  PushAL
  mov al,[ePCL]
  PushAL
  eConFlags              ;
  mov al,dl              ;
  cmp [ePE],1            ;
  jne SHORT @@IRQS       ;
  and al,11101111b       ;
@@IRQS:                     ;
  PushAL                 ;Push flags
  mov ah,0               ;
  mov bx,0fffeh          ;
  cmp [ePE],1
  je IRQS2
  mov bx,0ffeeh
  IRQS2:
  call MemRead           ;Get Reset Vector
  mov [ePB],0            ;clear program bank
  mov [ePCL],al          ;get reset vector AAL
  inc bx                 ;
  call MemRead           ;
  mov [ePCH],al          ;get reset vector AAH
  ret
ENDP


eCOP PROC NEAR
  mov al,[ePB]
  PushAL
  mov al,[ePCH]
  PushAL
  mov al,[ePCL]
  PushAL
  eConFlags              ;
  mov al,dl              ;
  cmp [ePE],1            ;
  jne SHORT @@COPS       ;
  and al,11101111b       ;
@@COPS:                     ;
  PushAL                 ;Push flags
  mov ah,0               ;
  mov bx,0fff4h          ;
  cmp [ePE],1
  je COPS2
  mov bx,0ffe4h
  COPS2:
  call MemRead           ;Get COP Vector
  mov [ePB],0            ;clear program bank
  mov [ePCL],al          ;get COP vector AAL
  inc bx                 ;
  call MemRead           ;
  mov [ePCH],al          ;get COP vector AAH
  ret
ENDP

eBRK PROC NEAR
  mov al,[ePB]
  PushAL
  mov al,[ePCH]
  PushAL
  mov al,[ePCL]
  PushAL
  eConFlags              ;
  mov al,dl              ;
  cmp [ePE],1            ;
  jne SHORT @@BRKS       ;
  and al,11101111b       ;
@@BRKS:                     ;
  PushAL                 ;Push flags
  mov ah,0               ;
  mov bx,0fffeh          ;
  cmp [ePE],1
  je BRKS2
  mov bx,0ffe6h
  BRKS2:
  call MemRead           ;Get COP Vector
  mov [ePB],0            ;clear program bank
  mov [ePCL],al          ;get COP vector AAL
  inc bx                 ;
  call MemRead           ;
  mov [ePCH],al          ;get COP vector AAH
  ret
ENDP


ResetCPU PROC NEAR
  mov [eDB],0            ;Reset Data Bank
  mov [eXH],0            ;Reset X-index high
  mov [eYH],0            ;Reset Y-index high
  mov [eSH],1            ;Reset Stack-pointer high
;  mov [eSL],0           ;
  mov [ePB],0            ;reset Program Bank
  mov [eDH],0            ;reset Direct Register
  mov [eDL],0            ;
  mov [ePm],20h          ;reset Flags
  mov [ePx],10h          ;
  mov [ePd],0            ;
  mov [ePi],4h           ;
  mov [ePE],1            ;
  mov al,[ePB]           ;
  PushAL                 ;Push Program Counter 
  mov al,[ePCH]          ;
  PushAL                 ;
  mov al,[ePCL]          ;
  PushAL                 ;
  eConFlags              ;
  mov al,dl              ;
  cmp [ePE],1            ;
  jne SHORT @@F          ;
  and al,11101111b       ;
@@F:                     ;
  PushAL                 ;Push flags
  mov ah,0               ;
  mov bx,0fffch          ;
  call MemRead           ;Get Reset Vector
  mov [ePB],0            ;clear program bank
  mov [ePCL],al          ;get reset vector AAL
  inc bx                 ;
  call MemRead           ;
  mov [ePCH],al          ;get reset vector AAH
  ret                    ;done with reset
ResetCPU ENDP

INCLUDE MEM.ASM
INCLUDE DEBUG.ASM
INCLUDE TWEAK.ASM
INCLUDE VIDEO.ASM
INCLUDE REG.ASM


  .FARDATA cSNESRom
  RBuf DB 32768 DUP (0)


END

